# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import ssl
from typing import Optional

from .types import tcpConnectionTypes
from .connection import TcpConnection, TcpConnectionUninitializedException
from ...common.types import HostPort, TcpOrTlsSocket
from ...common.utils import new_socket_connection
from ...common.constants import DEFAULT_SSL_CONTEXT_OPTIONS


class TcpServerConnection(TcpConnection):
    """A buffered server connection object."""

    def __init__(self, host: str, port: int) -> None:
        super().__init__(tcpConnectionTypes.SERVER)
        self._conn: Optional[TcpOrTlsSocket] = None
        self.addr: HostPort = (host, port)
        self.closed = True

    @property
    def connection(self) -> TcpOrTlsSocket:
        if self._conn is None:
            raise TcpConnectionUninitializedException()
        return self._conn

    def connect(
            self,
            addr: Optional[HostPort] = None,
            source_address: Optional[HostPort] = None,
    ) -> None:
        assert self._conn is None
        self._conn = new_socket_connection(
            addr or self.addr, source_address=source_address,
        )
        self.closed = False

    def wrap(
            self,
            hostname: Optional[str] = None,
            ca_file: Optional[str] = None,
            as_non_blocking: bool = False,
            # Ref https://github.com/PyCQA/pylint/issues/3691
            verify_mode: ssl.VerifyMode = ssl.VerifyMode.CERT_REQUIRED,     # pylint: disable=E1101
    ) -> None:
        ctx = ssl.create_default_context(ssl.Purpose.SERVER_AUTH, cafile=ca_file)
        ctx.options |= DEFAULT_SSL_CONTEXT_OPTIONS
        # pylint: disable=E1101
        ctx.check_hostname = (
            False if verify_mode == ssl.VerifyMode.CERT_NONE else hostname is not None
        )
        ctx.verify_mode = verify_mode
        self.connection.setblocking(True)
        self._conn = ctx.wrap_socket(
            self.connection,
            server_hostname=hostname,
        )
        if as_non_blocking:
            self.connection.setblocking(False)



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.

    Version definition.
"""
from typing import Tuple, Union


def _get_dist(distribution_name: str) -> str:
    # pylint: disable=import-outside-toplevel
    import warnings

    try:
        # pylint: disable=import-outside-toplevel
        from importlib.metadata import version  # noqa: WPS433

        return version(distribution_name)
    except ModuleNotFoundError:  # pragma: no cover
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', category=DeprecationWarning)

            # pylint: disable=import-outside-toplevel
            from pkg_resources import get_distribution  # noqa: WPS433

            return get_distribution(distribution_name).version


try:
    # pylint: disable=unused-import
    from ._scm_version import version as __version__  # noqa: WPS433, WPS436
    from ._scm_version import version_tuple as _ver_tup  # noqa: WPS433, WPS436
except ImportError:  # pragma: no cover
    __version__ = _get_dist('proxy.py')  # noqa: WPS440


def _to_int_or_str(inp: str) -> Union[int, str]:    # pragma: no cover
    try:
        return int(inp)
    except ValueError:
        return inp


def _split_version_parts(inp: str) -> Tuple[str, ...]:  # pragma: no cover
    public_version, _plus, local_version = inp.partition('+')
    return (*public_version.split('.'), local_version)


try:
    VERSION = _ver_tup
except NameError:   # pragma: no cover
    VERSION = tuple(
        map(_to_int_or_str, _split_version_parts(__version__)),
    )


__all__ = '__version__', 'VERSION'



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import os
import gzip
import logging
import multiprocessing
from typing import Any, Dict, Optional

from .base import BaseCacheResponsesPlugin
from ...http import httpStatusCodes
from .store.disk import OnDiskCacheStore
from ...http.parser import HttpParser, httpParserTypes
from ...common.constants import SLASH


br_installed = False
try:
    import brotli
    br_installed = True
except ModuleNotFoundError:
    pass

logger = logging.getLogger(__name__)


class CacheResponsesPlugin(BaseCacheResponsesPlugin):
    """Caches response using OnDiskCacheStore."""

    # Dynamically enable / disable cache
    ENABLED = multiprocessing.Event()

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.disk_store = OnDiskCacheStore(
            uid=self.uid,
            cache_dir=os.path.join(
                self.flags.cache_dir,
                'responses',
            ),
            cache_requests=self.flags.cache_requests,
        )
        self.set_store(self.disk_store)

    def on_access_log(self, context: Dict[str, Any]) -> Optional[Dict[str, Any]]:
        context.update({
            'cache_file_path': self.disk_store.cache_file_path,
        })
        return super().on_access_log(context)

    def on_upstream_connection_close(self) -> None:
        super().on_upstream_connection_close()
        if self.flags.cache_by_content_type and \
                self.disk_store.cache_file_path and \
                self.disk_store.cache_file_name:
            try:
                self.write_content_type(
                    self.disk_store.cache_file_path,
                    self.flags.cache_dir,
                    self.disk_store.cache_file_name,
                    self.flags.cache_requests,
                )
            except Exception as e:
                logger.exception('Unable to cache by content type', exc_info=e)

    @staticmethod
    def write_content_type(
            cache_file_path: str,
            cache_dir: str,
            content_file_name: str,
            cache_requests: bool,
    ) -> Optional[str]:
        if not cache_requests:
            parser = HttpParser(httpParserTypes.RESPONSE_PARSER)
            with open(cache_file_path, 'rb') as cache:
                data = cache.read()
                parser.parse(memoryview(data))
            if parser.code and int(parser.code) == httpStatusCodes.SWITCHING_PROTOCOLS:
                logger.warning('Bypassing websocket response packet')
                return None
            # Certain responses may never reach complete state
            #
            # E.g. Facebook sends HTML page with content-length: 0
            # and as chunked encoded, while keeping the connection alive.
            # This connection gets dropped after timeout period.
            # Hence, technically response never completed or rather
            # terminated cleanly with a CRLF tokens.
            #
            # E.g. When it's a web socket upgrade response, same problem
            # can arise.  The response may not reach completion state.
            #
            # if not parser.is_complete:
            #     logger.warning(data)
            #     return None
            if parser.body_expected and parser.body:
                body = parser.body
                if parser.has_header(b'content-encoding'):
                    encoding = parser.header(b'content-encoding')
                    if encoding == b'gzip':
                        body = gzip.decompress(body)
                    elif encoding == b'br' and br_installed:
                        body = brotli.decompress(body)
                    else:
                        logger.warning(
                            'Unsupported content encoding %s',
                            encoding,
                        )
                        return None
                content_type = parser.header(b'content-type') \
                    if parser.has_header(b'content-type') \
                    else b'text/plain'
                extension = content_type.split(
                    SLASH, maxsplit=1,
                )[-1].split(b';', maxsplit=1)[0].split(b'+', maxsplit=1)[0].decode('utf-8')
                content_file_path = os.path.join(
                    cache_dir, 'content',
                    '%s.%s' % (content_file_name, extension),
                )
                with open(content_file_path, 'wb') as content:
                    content.write(body)
                logger.info('Cached content file at %s', content_file_path)
                return content_file_path
        else:
            # Last dumped packet is likely the response
            # packet
            pass
        return None



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.

    .. spelling::

       http
       Submodules
       websocket
       Websocket
"""
from .frame import WebsocketFrame, websocketOpcodes
from .client import WebsocketClient
from .plugin import WebSocketTransportBasePlugin


__all__ = [
    'websocketOpcodes',
    'WebsocketFrame',
    'WebsocketClient',
    'WebSocketTransportBasePlugin',
]



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import os
import time
import threading
from typing import Any, Dict, Optional
from multiprocessing import connection

from .names import eventNames
from ...common.types import DictQueueType


class EventQueue:
    """Global event queue.  Must be a multiprocess safe queue capable of
    transporting other queues.  This is necessary because currently
    subscribers use a separate subscription queue to consume events.
    Subscription queue is exchanged over the global event queue.

    Each published event contains following schema::

        {
            'request_id': 'Globally unique request ID',
            'process_id': 'Process ID of event publisher. This '
                          'will be the process ID of acceptor workers.',
            'thread_id': 'Thread ID of event publisher. '
                         'When --threadless is enabled, this value '
                         'will be same for all the requests.'
            'event_timestamp': 'Time when this event occured',
            'event_name': 'one of the pre-defined or custom event name',
            'event_payload': 'Optional data associated with the event',
            'publisher_id': 'Optional publisher entity unique name',
        }

    """

    def __init__(self, queue: DictQueueType) -> None:
        self.queue = queue

    def publish(
        self,
        request_id: str,
        event_name: int,
        event_payload: Dict[str, Any],
        publisher_id: Optional[str] = None,
    ) -> None:
        self.queue.put({
            'process_id': os.getpid(),
            'thread_id': threading.get_ident(),
            'event_timestamp': time.time(),
            'request_id': request_id,
            'event_name': event_name,
            'event_payload': event_payload,
            'publisher_id': publisher_id,
        })

    def subscribe(
            self,
            sub_id: str,
            channel: connection.Connection,
    ) -> None:
        """Subscribe to global events.

        sub_id is a subscription identifier which must be globally
        unique.  channel MUST be a multiprocessing connection.
        """
        self.queue.put({
            'event_name': eventNames.SUBSCRIBE,
            'event_payload': {'sub_id': sub_id, 'conn': channel},
        })

    def unsubscribe(
            self,
            sub_id: str,
    ) -> None:
        """Unsubscribe by subscriber id."""
        self.queue.put({
            'event_name': eventNames.UNSUBSCRIBE,
            'event_payload': {'sub_id': sub_id},
        })



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import base64
import random
import logging
import ipaddress
from typing import Any, Dict, List, Optional

from ..http import Url, httpHeaders, httpMethods
from ..core.base import TcpUpstreamConnectionHandler
from ..http.proxy import HttpProxyBasePlugin
from ..common.flag import flags
from ..http.parser import HttpParser
from ..common.utils import text_, bytes_
from ..http.exception import HttpProtocolException
from ..common.constants import (
    COLON, ANY_INTERFACE_HOSTNAMES, LOCAL_INTERFACE_HOSTNAMES,
)


logger = logging.getLogger(__name__)

DEFAULT_HTTP_ACCESS_LOG_FORMAT = '{client_ip}:{client_port} - ' + \
    '{request_method} {server_host}:{server_port}{request_path} -> ' + \
    '{upstream_proxy_host}:{upstream_proxy_port} - ' + \
    '{response_code} {response_reason} - {response_bytes} bytes - ' + \
    '{connection_time_ms} ms'

DEFAULT_HTTPS_ACCESS_LOG_FORMAT = '{client_ip}:{client_port} - ' + \
    '{request_method} {server_host}:{server_port} -> ' + \
    '{upstream_proxy_host}:{upstream_proxy_port} - ' + \
    '{response_bytes} bytes - {connection_time_ms} ms'

# Run two separate instances of proxy.py
# on port 9000 and 9001 BUT WITHOUT ProxyPool plugin
# to avoid infinite loops.
DEFAULT_PROXY_POOL: List[str] = [
    # Yes you may use the instance running with ProxyPoolPlugin itself.
    # ProxyPool plugin will act as a no-op.
    # 'localhost:8899',
    #
    # Remote proxies
    # 'localhost:9000',
    # 'localhost:9001',
]

flags.add_argument(
    '--proxy-pool',
    action='append',
    nargs=1,
    default=DEFAULT_PROXY_POOL,
    help='List of upstream proxies to use in the pool',
)


class ProxyPoolPlugin(TcpUpstreamConnectionHandler, HttpProxyBasePlugin):
    """Proxy pool plugin simply acts as a proxy adapter for proxy.py itself.

    Imagine this plugin as setting up proxy settings for proxy.py instance itself.
    All incoming client requests are proxied to configured upstream proxies."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._endpoint: Url = self._select_proxy()
        # Cached attributes to be used during access log override
        self._metadata: List[Any] = [
            None, None, None, None,
        ]

    def handle_upstream_data(self, raw: memoryview) -> None:
        self.client.queue(raw)

    def before_upstream_connection(
            self, request: HttpParser,
    ) -> Optional[HttpParser]:
        """Avoids establishing the default connection to upstream server
        by returning None.

        TODO(abhinavsingh): Ideally connection to upstream proxy endpoints
        must be bootstrapped within it's own re-usable and garbage collected pool,
        to avoid establishing a new upstream proxy connection for each client request.

        See :class:`~proxy.core.connection.pool.UpstreamConnectionPool` which is a work
        in progress for SSL cache handling.
        """
        # We don't want to send private IP requests to remote proxies
        try:
            if ipaddress.ip_address(text_(request.host)).is_private:
                return request
        except ValueError:
            pass
        # If chosen proxy is the local instance, bypass upstream proxies
        assert self._endpoint.port and self._endpoint.hostname
        if self._endpoint.port == self.flags.port and \
                self._endpoint.hostname in LOCAL_INTERFACE_HOSTNAMES + ANY_INTERFACE_HOSTNAMES:
            return request
        # Establish connection to chosen upstream proxy
        endpoint_tuple = (text_(self._endpoint.hostname), self._endpoint.port)
        logger.debug('Using endpoint: {0}:{1}'.format(*endpoint_tuple))
        self.initialize_upstream(*endpoint_tuple)
        assert self.upstream
        try:
            self.upstream.connect()
        except TimeoutError:
            raise HttpProtocolException(
                'Timed out connecting to upstream proxy {0}:{1}'.format(
                    *endpoint_tuple,
                ),
            )
        except ConnectionRefusedError:
            # TODO(abhinavsingh): Try another choice, when all (or max configured) choices have
            # exhausted, retry for configured number of times before giving up.
            #
            # Failing upstream proxies, must be removed from the pool temporarily.
            # A periodic health check must put them back in the pool.  This can be achieved
            # using a data structure without having to spawn separate thread/process for health
            # check.
            raise HttpProtocolException(
                'Connection refused by upstream proxy {0}:{1}'.format(
                    *endpoint_tuple,
                ),
            )
        logger.debug(
            'Established connection to upstream proxy {0}:{1}'.format(
                *endpoint_tuple,
            ),
        )
        return None

    def handle_client_request(
            self, request: HttpParser,
    ) -> Optional[HttpParser]:
        """Only invoked once after client original proxy request has been received completely."""
        if not self.upstream:
            return request
        assert self.upstream
        # For log sanity (i.e. to avoid None:None), expose upstream host:port from headers
        host, port = None, None
        # Browser or applications may sometime send
        #
        # "CONNECT / HTTP/1.0\r\n\r\n"
        #
        # for proxy keep alive checks.
        if request.has_header(b'host'):
            url = Url.from_bytes(request.header(b'host'))
            assert url.hostname
            host, port = url.hostname.decode('utf-8'), url.port
            port = port if port else (
                443 if request.is_https_tunnel else 80
            )
        path = None if not request.path else request.path.decode()
        self._metadata = [
            host, port, path, request.method,
        ]
        # Queue original request optionally with auth headers to upstream proxy
        if self._endpoint.has_credentials:
            assert self._endpoint.username and self._endpoint.password
            request.add_header(
                httpHeaders.PROXY_AUTHORIZATION,
                b'Basic ' +
                base64.b64encode(
                    self._endpoint.username +
                    COLON +
                    self._endpoint.password,
                ),
            )
        self.upstream.queue(memoryview(request.build(for_proxy=True)))
        return request

    def handle_client_data(self, raw: memoryview) -> Optional[memoryview]:
        """Only invoked when before_upstream_connection returns None"""
        # Queue data to the proxy endpoint
        assert self.upstream
        self.upstream.queue(raw)
        return raw

    def handle_upstream_chunk(self, chunk: memoryview) -> Optional[memoryview]:
        """Will never be called since we didn't establish an upstream connection."""
        if not self.upstream:
            return chunk
        # pylint: disable=broad-exception-raised
        raise Exception("This should have never been called")

    def on_upstream_connection_close(self) -> None:
        """Called when client connection has been closed."""
        if self.upstream and not self.upstream.closed:
            logger.debug('Closing upstream proxy connection')
            self.upstream.close()
            self.upstream = None

    def on_access_log(self, context: Dict[str, Any]) -> Optional[Dict[str, Any]]:
        if not self.upstream:
            return context
        addr, port = (self.upstream.addr[0], self.upstream.addr[1]) \
            if self.upstream else (None, None)
        context.update({
            'upstream_proxy_host': addr,
            'upstream_proxy_port': port,
            'server_host': self._metadata[0],
            'server_port': self._metadata[1],
            'request_path': self._metadata[2],
            'response_bytes': self.total_size,
        })
        self.access_log(context)
        return None

    def access_log(self, log_attrs: Dict[str, Any]) -> None:
        access_log_format = DEFAULT_HTTPS_ACCESS_LOG_FORMAT
        request_method = self._metadata[3]
        if request_method and request_method != httpMethods.CONNECT:
            access_log_format = DEFAULT_HTTP_ACCESS_LOG_FORMAT
        logger.info(access_log_format.format_map(log_attrs))

    def _select_proxy(self) -> Url:
        """Choose a random proxy from the pool.

        TODO: Implement your own logic here e.g. round-robin, least connection etc.
        """
        return Url.from_bytes(bytes_(random.choice(self.flags.proxy_pool)[0]))



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
from typing import Any, List, Tuple, Generator, cast
from urllib.parse import parse_qs, urlparse
from multiprocessing.synchronize import Lock

from .plugin import HttpWebServerBasePlugin
from ..parser import HttpParser
from .protocols import httpProtocolTypes
from ...common.flag import flags
from ...common.utils import text_, build_http_response
from ...common.constants import (
    DEFAULT_ENABLE_METRICS, DEFAULT_METRICS_URL_PATH,
)
from ...core.event.metrics import MetricsStorage


flags.add_argument(
    '--enable-metrics',
    action='store_true',
    default=DEFAULT_ENABLE_METRICS,
    help='Default: False.  Enables metrics.',
)

flags.add_argument(
    '--metrics-path',
    type=str,
    default=text_(DEFAULT_METRICS_URL_PATH),
    help='Default: %s. Web server path to serve proxy.py metrics.'
    % text_(DEFAULT_METRICS_URL_PATH),
)


def get_collector(metrics_lock: Lock) -> Any:
    # pylint: disable=import-outside-toplevel
    from prometheus_client.core import Metric
    from prometheus_client.registry import Collector

    class MetricsCollector(Collector):

        def __init__(self, metrics_lock: Lock) -> None:
            self.storage = MetricsStorage(metrics_lock)

        def collect(self) -> Generator[Metric, None, None]:
            """Serves from aggregates metrics managed by MetricsEventSubscriber."""
            # pylint: disable=import-outside-toplevel
            from prometheus_client.core import (
                GaugeMetricFamily, CounterMetricFamily,
            )

            started = self.storage.get_counter('work_started')
            finished = self.storage.get_counter('work_finished')

            work_started = CounterMetricFamily(
                'proxypy_work_started',
                'Total work accepted and started by proxy.py core',
            )
            work_started.add_metric(
                ['proxypy_work_started'],
                started,
            )
            yield work_started

            request_complete = CounterMetricFamily(
                'proxypy_work_request_received',
                'Total work finished sending initial request',
            )
            request_complete.add_metric(
                ['proxypy_work_request_received'],
                self.storage.get_counter('request_complete'),
            )
            yield request_complete

            work_finished = CounterMetricFamily(
                'proxypy_work_finished',
                'Total work finished by proxy.py core',
            )
            work_finished.add_metric(
                ['work_finished'],
                finished,
            )
            yield work_finished

            ongoing_work = GaugeMetricFamily(
                'proxypy_work_active',
                'Total work under active execution',
                value=started - finished,
            )
            yield ongoing_work

    return MetricsCollector(metrics_lock)


class MetricsWebServerPlugin(HttpWebServerBasePlugin):

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        # pylint: disable=import-outside-toplevel
        from prometheus_client.core import CollectorRegistry
        from prometheus_client.registry import Collector

        self.registry = CollectorRegistry()
        self.registry.register(cast(Collector, get_collector(self.flags.metrics_lock)))

    def routes(self) -> List[Tuple[int, str]]:
        if self.flags.metrics_path:
            return [
                (
                    httpProtocolTypes.HTTP,
                    r'{0}$'.format(
                        text_(self.flags.metrics_path),
                    ),
                ),
                (
                    httpProtocolTypes.HTTPS,
                    r'{0}$'.format(
                        text_(self.flags.metrics_path),
                    ),
                ),
            ]
        return []  # pragma: no cover

    def handle_request(self, request: HttpParser) -> None:
        # pylint: disable=import-outside-toplevel
        from prometheus_client.exposition import _bake_output

        # flake8: noqa
        status, headers, output = _bake_output(  # type: ignore[no-untyped-call]
            self.registry,
            (
                request.header(b'Accept').decode()
                if request.has_header(b'Accept')
                else '*/*'
            ),
            (
                request.header(b'Accept-Encoding').decode()
                if request.has_header(b'Accept-Encoding')
                else None
            ),
            parse_qs(urlparse(request.path).query),
            False,
        )
        statuses = status.split(' ', maxsplit=1)
        response = build_http_response(
            int(statuses[0]),
            reason=statuses[1].encode(),
            headers={key.encode(): value.encode() for key, value in headers},
            body=output,
        )
        self.client.queue(memoryview(response))



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
from typing import Optional

from ..http.proxy import HttpProxyBasePlugin
from ..http.parser import HttpParser
from ..common.utils import bytes_
from ..common.version import __version__


class ModifyRequestHeaderPlugin(HttpProxyBasePlugin):
    """Modify request header before sending to upstream server."""

    # def before_upstream_connection(self, request: HttpParser) -> Optional[HttpParser]:
    #     """NOTE: Use this for HTTP only request headers modification."""
    #     request.add_header(
    #         b"x-proxy-py-version",
    #         bytes_(__version__),
    #     )
    #     return request

    def handle_client_request(self, request: HttpParser) -> Optional[HttpParser]:
        """NOTE: This is for HTTPS request headers modification when under TLS interception.

        For HTTPS requests, modification of request under TLS interception WILL NOT WORK
        through before_upstream_connection.
        """
        request.add_header(
            b'x-proxy-py-version',
            bytes_(__version__),
        )
        return request



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.

    .. spelling::

       auth
       http
"""
from typing import TYPE_CHECKING, Any

from .base import HttpProtocolException
from ..responses import PROXY_AUTH_FAILED_RESPONSE_PKT


if TYPE_CHECKING:   # pragma: no cover
    from ..parser import HttpParser


class ProxyAuthenticationFailed(HttpProtocolException):
    """Exception raised when HTTP Proxy auth is enabled and
    incoming request doesn't present necessary credentials."""

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(self.__class__.__qualname__, **kwargs)

    def response(self, _request: 'HttpParser') -> memoryview:
        return PROXY_AUTH_FAILED_RESPONSE_PKT
